#include "mbed.h"

#define  NETIF 1

#if NETIF

#include "EthernetNetIf.h"
#include "SimpleSocket.h"

#else

#include <Sockets.h>
#include <mbedNetIF.h>
#include <ARP.h>
#include <ICMPv4.h>
#include <UDPv4.h>
#include <Debug.h>

#endif

#include "CommDefs.h"
#include "EncoderDefs.h"

#define NET_TIMER_PERIOD        1
#define BUFFER_LEN              1024

#define DEBUGGING 0

#define MY_VERNO 1.031

Serial pc( USBTX, USBRX );

// Currently Used Pins:
//
//    p11, p12 - Encoder 1
//    p13, p14 - Encoder 2
//    p15, p16 - Encoder 3
//    p17, p18 - Encoder 4
//    p19, p20 - Encoder 5
//    p21, p22 - Encoder 6
//    p23, p24 - Encoder 7
//    p25, p26 - Encoder 8
//
//    p27      - Status LED
//    p28      - Configuration Mode Button
//    
//    p30 - Green LED Ethernet
//    p29 - Yellow LED Ethernet
//    p33-36   - Ethernet
//   


DigitalOut StatusLED( p27 );
DigitalIn  ConfigModeButton( p28 );
DigitalOut GreenEthLED( p30 );
DigitalOut YellowEthLED( p29 );

long ConfigMode = 0;
LocalFileSystem local( "local" );

long FLASHCOUNT = 0;

unsigned long BOARDID  = 0x0001;
uint8_t       IPStr[4]   = { 192, 168,   1, 252 };  // My Default IP Address
uint8_t       NMStr[4]   = { 255, 255, 255,   0 };  // My NetMask
uint8_t       GWStr[4]   = { 192, 168,   1,   1 };  // My Gateway
uint8_t       HostIp[4]  = { 192, 168,   1,   1 };  // My Host
int           IPort = 6100;

float SendTime = 0.05;
float RecvTime = 0.1;

// Time-Based Inteerupt
Ticker ticker;
unsigned long WRDATA  = false;

TOENC   *pToEncoder;
FROMENC FromEncoder;

//Host buddy;
char buff[4096];
char locbuff[4096];

int  BuffLen;

long TimeoutCount = 0;
long TimeoutFlag = false;

Ticker              netTicker;

#if NETIF

Host buddy;

#else

char                MACAddr[10];
NetIF_t             *mbedNetIF;
Socket_t            udpSocket;
Socket_AddrIn_t     localAddress, senderAddress, broadcastAddress;


void NetTicker(void)
{
    NetIF_ProcessTimers( NET_TIMER_PERIOD );
}

#endif

void EncoderHandler()
{
    int ii, jj;
    
    TimeoutCount++;

    ENCLOOP++;    
    for ( ii = 0; ii < 8; ii++ )
    {    
        EncRaw[ii] = ( EncRaw[ii] << 2 ) + ( EncPairA[ii] << 1 ) + EncPairB[ii];
        jj = EncRaw[ii] &= 0x0F;
        EncVal[ii] += EncResponse[EncInc[jj]]; 
      
        // If the encoder is unknown increment in the same direction as previous tick
        if ( jj == ENC_UNKNOWN ) EncVal[ii] += 2 * LastInc;
        else                     LastInc = EncResponse[EncInc[jj]];
         
        if ( EncType[ii] == ENC_ROLLOVER )
        {
            if ( EncVal[ii] > EncMax[ii] ) EncVal[ii] = EncMin[ii];
            if ( EncVal[ii] < EncMin[ii] ) EncVal[ii] = EncMax[ii];
        }
        else if ( EncType[ii] == ENC_HARDSTOP )
        {
            if ( EncVal[ii] > EncMax[ii] ) EncVal[ii] = EncMax[ii];
            if ( EncVal[ii] < EncMin[ii] ) EncVal[ii] = EncMin[ii];
        };
    };
    
    if ( FLASHCOUNT++ > 10000 ) FLASHCOUNT = 0;
    StatusLED = ( ( FLASHCOUNT < 5000 ) || !ConfigMode );
};

void InitPins( void )
{
    ticker.attach_us( EncoderHandler, 100 ); 
    ConfigModeButton.mode( PullUp );
};

void InitStructures( void )
{
    FromEncoder.HEADER  = FROMENCHEADER;
    FromEncoder.TRAILER = FROMENCTRAILER; 
    FromEncoder.BOARDID = BOARDID;
    FromEncoder.PACKTYPE = 0x02;     
};

long ReadDef( void )
{
    FILE *Stream = fopen( "/local/Encoder.def", "rb" );    
    if ( Stream )
    {
       
        fread( (char *)&BOARDID,    sizeof( BOARDID ),  1, Stream );
        fread( (char *)&IPStr[0],   sizeof( IPStr ),    1, Stream );
        fread( (char *)&NMStr[0],   sizeof( NMStr ),    1, Stream );
        fread( (char *)&GWStr[0],   sizeof( GWStr ),    1, Stream );
        fread( (char *)&HostIp[0],  sizeof( HostIp ),   1, Stream );
        fread( (char *)&IPort,      sizeof( IPort ),    1, Stream );
        fread( (char *)&EncMin[0],  sizeof( EncMin ),   1, Stream );
        fread( (char *)&EncMax[0],  sizeof( EncMax ),   1, Stream );
        fread( (char *)&SendTime,   sizeof( SendTime ), 1, Stream );
        fread( (char *)&RecvTime,   sizeof( RecvTime ), 1, Stream );
        fclose( Stream );
        return true;
    }
    else
    {
        return false;
    };
 };  

long WriteDef( void )
{
    FILE *Stream = fopen( "/local/Encoder.def", "wb" );
    if ( Stream )
    {
        fwrite( (char *)&BOARDID,    sizeof( BOARDID ),  1, Stream );
        fwrite( (char *)&IPStr[0],   sizeof( IPStr ),    1, Stream );
        fwrite( (char *)&NMStr[0],   sizeof( NMStr ),    1, Stream );
        fwrite( (char *)&GWStr[0],   sizeof( GWStr ),    1, Stream );
        fwrite( (char *)&HostIp[0],  sizeof( HostIp ),   1, Stream );
        fwrite( (char *)&IPort,      sizeof( IPort ),    1, Stream );
        fwrite( (char *)&EncMin[0], sizeof( EncMin ),    1, Stream );
        fwrite( (char *)&EncMax[0], sizeof( EncMax ),    1, Stream );
        fwrite( (char *)&SendTime,   sizeof( SendTime ), 1, Stream );
        fwrite( (char *)&RecvTime,   sizeof( RecvTime ), 1, Stream );
        fclose( Stream );
        return true;
    }
    else
    {
        return false;
    };
};

int ProcessPacks( void )
{
    int ii;
    
    // Only Process Valid Packets
    if ( ( BuffLen == sizeof( TOENC ) ) 
      && ( pToEncoder->HEADER  == TOENCHEADER )
      && ( pToEncoder->TRAILER == TOENCTRAILER ) 
      && ( pToEncoder->BOARDID == BOARDID ) )
    {
    
        pToEncoder->HEADER = 0;
        FromEncoder.STATUS[9] = ENCLOOP;                          
        
        for ( ii = 0; ii < 8; ii++ )
        {
            FromEncoder.ENCODERS[ii] = EncVal[ii]; 
        };
        
        if ( pToEncoder->PACKTYPE == 0x0001 )  // Normal Pack
        {
            // Send packet back
            FromEncoder.PACKTYPE = 0x0002;
            
            for ( ii = 0; ii < 8; ii++ )
            {
                if ( pToEncoder->CWORD[ii] & SET_ENCODER )
                {
                    EncVal[ii] = pToEncoder->ENC_SET[ii]; 
                };
                
                FromEncoder.ENCODERS[ii] = EncVal[ii]; 
            };
            
        }
        else if ( pToEncoder->PACKTYPE == 0x0100 ) // Configuration Pack
        {
            printf( "C!" );
            
            // Send packet back                        
            FromEncoder.PACKTYPE = 0x0200;
            
            BOARDID = pToEncoder->BOARDID;
            FromEncoder.BOARDID = BOARDID; 
            
            IPStr[0] = pToEncoder->ENCIP[0];
            IPStr[1] = pToEncoder->ENCIP[1];
            IPStr[2] = pToEncoder->ENCIP[2];
            IPStr[3] = pToEncoder->ENCIP[3];
            
            NMStr[0] = pToEncoder->NETMASK[0];
            NMStr[1] = pToEncoder->NETMASK[1];
            NMStr[2] = pToEncoder->NETMASK[2];
            NMStr[3] = pToEncoder->NETMASK[3];
            
            GWStr[0] = pToEncoder->GATEWAY[0];
            GWStr[1] = pToEncoder->GATEWAY[1];
            GWStr[2] = pToEncoder->GATEWAY[2];
            GWStr[3] = pToEncoder->GATEWAY[3];
            
            HostIp[0] = pToEncoder->HOSTIP[0];
            HostIp[1] = pToEncoder->HOSTIP[1];
            HostIp[2] = pToEncoder->HOSTIP[2];
            HostIp[3] = pToEncoder->HOSTIP[3];
            
            IPort    = pToEncoder->PORT;
            
            SendTime = pToEncoder->SendTime;
            RecvTime = pToEncoder->RecvTime;
            
            for ( ii = 0; ii < 8; ii++ )
            {
                if ( pToEncoder->CWORD[ii] & SET_ENCODER )
                {
                    EncVal[ii] = pToEncoder->ENC_SET[ii]; 
                };
                
                EncType[ii] = pToEncoder->EncType[ii];
                EncMin[ii]  = pToEncoder->EncMin[ii];
                EncMax[ii]  = pToEncoder->EncMax[ii];
            };
            
            WRDATA = true;
        };                        
        
        return true;
        
    };
    
    return false;
};

int main() 
{
    Timer t;
    Timer SendBuff; 
    Timer CheckBuff;
    
    pc.baud( 115200 );
    
    // Some Initialization Of Pins For Encoder
    pc.printf( "Software Version Number %f\n Initializing\n", MY_VERNO );    
    InitPins();
    wait( 0.1 );
    
    ConfigMode = ConfigModeButton;
    
    // Look at definition file
    if ( ConfigMode )
    {
        pc.printf( "Using Defaults For Configuration Mode!\n" );
    }
    else
    {
        pc.printf( "Reading Data File For Initialization and Normal Operation!\n" ); 
        ReadDef();
    };
    
    if ( ( SendTime <= 0.0 ) || ( RecvTime <= 0.0 ) )
    {
        SendTime = 0.02;
        RecvTime = 0.1;
        pc.printf( "Problem with config! Using Defaults times\n" );
    };
    
    pc.printf( "Send Frequency: %0.3fHz Recv Frequency: %0.3fHz\n", ( 1.0 / SendTime ), ( 1.0 / RecvTime ) ); 
    
    InitStructures();
    
    pc.printf( "Setting Up Datagram\n" );   
    
#if NETIF

    // Setup Ethernet Interface
    pc.printf( "Setting IP Address to [%d.%d.%d.%d] Port: %d \n", IPStr[0], IPStr[1], IPStr[2], IPStr[3], IPort );
    EthernetNetIf eth( IpAddr( IPStr[0], IPStr[1], IPStr[2], IPStr[3] ),
                       IpAddr( NMStr[0], NMStr[1], NMStr[2], NMStr[3] ),
                       IpAddr( GWStr[0], GWStr[1], GWStr[2], GWStr[3] ),
                       IpAddr( GWStr[0], GWStr[1], GWStr[2], GWStr[3] )   );

    EthernetErr ethErr = eth.setup( 15000 );

    if( ethErr )
    {    
        pc.printf( "Could not setup Ethernet... Halting Application: Err #%d\n", ethErr );
        return -1;
    }

    pc.printf( "Setting Up Datagram\n" );
    
    
    DatagramSocket datagram( IPort, DEBUGGING );


#else
            
    IPv4_Addr_t         ip =            { IPStr[0], IPStr[1], IPStr[2], IPStr[3] },
                        netmask =       { NMStr[0], NMStr[1], NMStr[2], NMStr[3] },
                        gateway =       { GWStr[0], GWStr[1], GWStr[2], GWStr[3] },
                        broadcastNet =  { 192, 168,   1, 255 },
                        broadcastAll =  { 255, 255, 255, 255 };
                        
    mbed_mac_address( MACAddr );
    Ethernet_Addr_t myMAC =  { MACAddr[0], MACAddr[1], MACAddr[2], 
                               MACAddr[3], MACAddr[4], MACAddr[5]  };
    

    // Set debug levels to have the NetIF networking layer display the interface status
    Debug_SetMasks( DEBUG_MODULE_NETIF, 
                    DEBUG_LEVEL_ALL & ~(DEBUG_LEVEL_VERBOSE0 | DEBUG_LEVEL_VERBOSE1 | DEBUG_LEVEL_VERBOSE2 ) );

    // Registers the ethernet drivers as a network interface
    mbedNetIF = NetIF_RegisterInterface( &ip, &netmask, &gateway, &mbedNetIF_Driver, (void *)&myMAC );

    // We need ARP over ethernet to resolve IP addresses with MAC addresses
    ethernet.RegisterProtocol( &arp );      
    
    // We need IPv4 (over ethernet) of course
    ethernet.RegisterProtocol( &ipv4 );     
    
    // Then ICMPv4 over IPv4 to answer ping requests
    ipv4.RegisterProtocol( &icmpv4 );       
    
    // Next let's register UDPv4 over IPv4
    ipv4.RegisterProtocol( &udpv4 );      
    
    // Finally let's register the Sockets API (it is available for UDP at the moment)  
    udpv4.RegisterAPI( &sockets );          

    // We add a static entry to the local network broadcast address to the ARP cache
    ARP_AddStaticEntry( mbedNetIF, broadcastNet, &ethernet_Addr_Broadcast ); 

    // We add a static entry to the GLOBAL broadcast address to the ARP cache
    ARP_AddStaticEntry( mbedNetIF, broadcastAll, &ethernet_Addr_Broadcast ); 

    // A 1 second periodic timer to process network house keeping routines
    //   (like expiration in the ARP cache, timeouts for TCP in the future, ...)
    netTicker.attach( &NetTicker, 0.003 ); // NET_TIMER_PERIOD );     
    
    // Finally we bring the network interface up
    NetIF_Up( mbedNetIF );                                
    
    udpSocket = Sockets_Open( AF_INET, SOCK_DGRAM, 0 );   // Create an UDP socket
    
    localAddress.family = AF_INET;
    localAddress.len = sizeof( Socket_AddrIn_t );
    localAddress.port = htons( IPort );
    localAddress.address = ip;
    
    broadcastAddress.family = AF_INET;
    broadcastAddress.len = sizeof(Socket_AddrIn_t);
    broadcastAddress.port = htons( IPort );
    broadcastAddress.address = broadcastNet;

    Sockets_Bind( udpSocket, (Socket_Addr_t *)&localAddress, localAddress.len );  // We bind the socket to port IPort and any local IP address

    // Then do whatever you need for your application. 
    // Here we simply check if we have received a datagram from the udpSocket and display its informations
    int32_t senderAddressLength = sizeof(Socket_AddrIn_t);

    // Setup Ethernet Interface
    pc.printf( "Setting IP Address to [%d.%d.%d.%d] Port: %d \n", IPStr[0], IPStr[1], IPStr[2], IPStr[3], IPort );
    
#endif
    
    GreenEthLED = true;

    pc.printf( "In Main Loop...\n" );        
    
    t.start();
    CheckBuff.start();
    SendBuff.start();
    while ( true ) 
    {
        if ( t.read() > 1.0 ) 
        {
            pc.printf( "T" );
            t.reset();
        };
        
        if ( ENCLOOP >= 10000 )
        {
            ENCLOOP = 0;
            pc.printf( "*" );
        };
        
        if ( ( TimeoutCount >= 10000 ) && !TimeoutFlag )
        {
            TimeoutFlag = true;
            TimeoutCount = 10000;
            pc.printf( "<Not Connected>\n" );            
        };
        
#if NETIF

        // Handle UDP Packets
        if ( datagram.receive( &buddy, 0.0 ) > 0 ) 
        {
            pc.printf( "R" );
        
            YellowEthLED = true;
            IpAddr ip   = buddy.getIp();
            int    port = buddy.getPort();
            BuffLen = datagram.read( buff, 200 );
            
            pToEncoder = (TOENC *)&buff;
            
 //           pc.printf( "Reading...%d %d 0x%08X 0x%08X 0x%08X", BuffLen, sizeof( TOENC ), pToEncoder->HEADER, pToEncoder->BOARDID, pToEncoder->TRAILER );
            if ( ProcessPacks() )
            {
                TimeoutCount = 0;
                if ( TimeoutFlag ) 
                {
                    pc.printf( "[Connected]\n" );                
                    TimeoutFlag = false;
                }
                
                pc.printf( "+" );
                datagram.write(  (char *)&FromEncoder, sizeof( FromEncoder ) );
                datagram.send( ip, port );                            
                pc.printf( "AW" );
            };       
            
 //           pc.printf( "\n" );
        }

#else
        
        if ( ( CheckBuff.read() > RecvTime ) && ( ConfigMode ) )
        {
            CheckBuff.reset();
            
            // Handle UDP Packets
            BuffLen = Sockets_RecvFrom( udpSocket, 
                                        (uint8_t *)buff, 
                                        BUFFER_LEN, 
                                        0, 
                                        (Socket_Addr_t *)&senderAddress, 
                                        &senderAddressLength );
                                        
            if ( BuffLen == sizeof( TOENC ) ) 
            {
                pc.printf( "R" );
            
                YellowEthLED = true;
                
                memcpy( locbuff, buff, BuffLen );
                pToEncoder = (TOENC *)&locbuff;
                
                if ( ProcessPacks() )
                {
                    pc.printf( "+" );
                    TimeoutCount = 0;
                    if ( TimeoutFlag ) 
                    {
                        pc.printf( "[Connected]\n" );                
                        TimeoutFlag = false;
                    }
                    
               };
            };   
        };        
                
        if ( SendBuff.read() > SendTime )
        {           
            // Send packet back
            FromEncoder.STATUS[9] = (unsigned long)( MY_VERNO * 1000.0 );
            for ( int ii = 0; ii < 8; ii++ )
            {               
                FromEncoder.ENCODERS[ii] = EncVal[ii]; 
            };
            
            Sockets_SendTo( udpSocket, (uint8_t *)&FromEncoder,  sizeof( FromEncoder ), 0, (Socket_Addr_t *)&broadcastAddress, broadcastAddress.len );
            pc.printf( "S" );
            
            SendBuff.reset();
        };          
#endif

        // Handle The Writing Of Data To A Config File
        if ( WRDATA )
        {
            pc.printf( "\n[Writing Data]\n" );
            WriteDef();           
            wait( 1.0 ); 
            WRDATA = false; 
            
        };
        
        YellowEthLED = false;
       
    }
}